from django.middleware.cache import UpdateCacheMiddleware
from django.utils.cache import patch_vary_headers
from phased.utils import second_pass_render, drop_vary_headers


class PhasedRenderMiddleware(object):
    def process_response(self, request, response):
        if not response['content-type'].startswith("text/html"):
            return response
        response.content = second_pass_render(request, response.content)
        response['Content-Length'] = str(len(response.content))
        return response


class PatchedVaryUpdateCacheMiddleware(UpdateCacheMiddleware):
    def process_response(self, request, response):
        # If "Vary: Cookie" is set in the response object, Django's cache
        # middleware will vary the cache key based on the value of the cookie.
        # This removes that header.
        drop_vary_headers(response, ['Cookie'])

        response = super(PatchedVaryUpdateCacheMiddleware, self).process_response(request, response)

        # This re-adds the header that was dropped above, so browsers still
        # know to vary on cookies.
        patch_vary_headers(response, ['Cookie'])

        return response



import re, base64
from django.template.context import BaseContext, RequestContext
from django.template import (Parser, Lexer, Token,
    TOKEN_TEXT, COMMENT_TAG_START, COMMENT_TAG_END, TemplateSyntaxError)
from django.utils.cache import cc_delim_re
from django.utils.functional import Promise, LazyObject
from django.http import HttpRequest
from django.contrib.messages.storage.base import BaseStorage

try:
    import cPickle as pickle
except ImportError:
    import pickle

from phased import settings

pickled_context_re = re.compile(r'.*%s stashed context: "(.*)" %s.*' % (COMMENT_TAG_START, COMMENT_TAG_END))
forbidden_classes = (Promise, LazyObject, HttpRequest, BaseStorage)

def second_pass_render(request, content):
    """
    Split on the literal delimiter and generate the token list by passing
    through text outside of literal blocks as single text tokens and tokenizing
    text inside the literal blocks. This ensures that nothing outside of the
    literal blocks is tokenized, thus eliminating the possibility of a template
    code injection vulnerability.
    """
    result = tokens = []
    for index, bit in enumerate(content.split(settings.LITERAL_DELIMITER)):
        if index % 2:
            tokens = Lexer(bit, None).tokenize()
        else:
            tokens.append(Token(TOKEN_TEXT, bit))
        context = RequestContext(request, unpickle_context(bit))
        result.append(Parser(tokens).parse().render(context))
    return "".join(result)

def drop_vary_headers(response, headers_to_drop):
    """
    Remove an item from the "Vary" header of an ``HttpResponse`` object.
    If no items remain, delete the "Vary" header.
    This does the opposite effect of django.utils.cache.patch_vary_headers.
    """
    if response.has_header('Vary'):
        vary_headers = cc_delim_re.split(response['Vary'])
    else:
        vary_headers = []

    headers_to_drop = [header.lower() for header in headers_to_drop]

    updated_vary_headers = []
    for header in vary_headers:
        if len(header):
            if header.lower() not in headers_to_drop:
                updated_vary_headers.append(header)

    if len(updated_vary_headers):
        response['Vary'] = ', '.join(updated_vary_headers)
    else:
        del response['Vary']

def flatten_context(context, remove_lazy=True):
    """
    Creates a dictionary from a Context instance by traversing
    its dicts list. Can remove unwanted subjects from the result,
    e.g. lazy objects.
    """
    flat_context = {}
    for context_dict in context.dicts:
        if remove_lazy:
            result = dict((k, v) for k, v in context_dict.iteritems() if not isinstance(v, forbidden_classes))
            flat_context.update(result)
        else:
            flat_context.update(context_dict)
    return flat_context

def unpickle_context(content, pattern=None):
    """
    Unpickle the context from the given content string or return None.
    """
    if pattern is None:
        pattern = pickled_context_re
    match = pattern.search(content)
    if match:
        return pickle.loads(base64.standard_b64decode(match.group(1)))
    return None

def pickle_context(context, template=None):
    """
    Pickle the given Context instance and do a few optimzations before.
    """
    if not isinstance(context, BaseContext):
        raise TemplateSyntaxError('Literal context is not a Context instance')
    pickled_context = pickle.dumps(flatten_context(context), protocol=pickle.HIGHEST_PROTOCOL)
    if template is None:
        template = '{# stashed context: "%s" #}'
    return template % base64.standard_b64encode(pickled_context)



from django.template import (Library, Node, resolve_variable,
    TOKEN_BLOCK, TOKEN_COMMENT, TOKEN_TEXT, TOKEN_VAR,
    TemplateSyntaxError, VariableDoesNotExist, Context)
from django.utils.encoding import smart_str

from phased import settings
from phased.utils import pickle_context

register = Library()

def parse(parser):
    """
    Parse to the end of a literal block. This is different than Parser.parse()
    in that it does not generate Node objects; it simply yields tokens.
    """
    depth = 0
    while parser.tokens:
        token = parser.next_token()
        if token.token_type == TOKEN_BLOCK:
            if token.contents == 'literal':
                depth += 1
            elif token.contents == 'endliteral':
                depth -= 1
        if depth < 0:
            break
        yield token
    if not parser.tokens and depth >= 0:
        parser.unclosed_block_tag(('endliteral',))


class LiteralNode(Node):
    def __init__(self, content, var_names):
        self.var_names = var_names
        self.content = content

    def __repr__(self):
        return "<Literal Node: '%s'>" % smart_str(self.content[:25], 'ascii',
                errors='replace')

    def render(self, context):
        stash = {}
        for var_name in self.var_names:
            if var_name[0] in ('"', "'") and var_name[-1] == var_name[0]:
                var_name = var_name[1:-1]
            try:
                stash[var_name] = resolve_variable(var_name, context)
            except VariableDoesNotExist:
                raise TemplateSyntaxError(
                    '"literal" tag got an unknown variable: %r' % var_name)
        pickled = None
        if not stash and settings.KEEP_CONTEXT:
            pickled = pickle_context(context)
        elif stash:
            pickled = pickle_context(Context(stash))
        return '%(delimiter)s%(content)s%(pickled)s%(delimiter)s' % {
            'delimiter': settings.LITERAL_DELIMITER,
            'content': self.content,
            'pickled': pickled or '',
        }


def do_literal(parser, token):
    """
    Template tag to denote a template section to render a second time via
    a middleware.

    Usage::

        {% load phased_tags %}
        {% literal with [var1] [var2] .. %}
            .. some content to be rendered a second time ..
        {% endliteral %}

    You can pass it a list of context variable names to automatically
    save those variables for the second pass rendering of the template,
    e.g.::

        {% load phased_tags %}
        {% literal with comment_count object %}
            There are {{ comment_count }} comments for "{{ object }}".
        {% endliteral %}

    Alternatively you can also set the ``PHASED_KEEP_CONTEXT`` setting to
    ``True`` to automatically keep the whole context for each literal block.

    Note: Lazy objects such as messages and csrf tokens aren't kept.

    """
    literal = ''.join({
        TOKEN_BLOCK: '{%% %s %%}',
        TOKEN_VAR: '{{ %s }}',
        TOKEN_COMMENT: '{# %s #}',
        TOKEN_TEXT: '%s',
    }[token.token_type] % token.contents for token in parse(parser))
    tokens = token.contents.split()
    if len(tokens) > 1 and tokens[1] != 'with':
        raise TemplateSyntaxError(u"'%r' tag requires the second argument to be 'with'." % tokens[0])
        if len(tokens) == 2:
            raise TemplateSyntaxError(u"'%r' tag requires at least one context variable name." % tokens[0])
    return LiteralNode(literal, tokens[2:])

register.tag('literal', do_literal)
